@@ -248,7 +248,10 @@ int main(int argc, char *argv[])
     sigaction(SIGQUIT, &act, NULL);
 
     memset(&hints, 0, sizeof(hints));
-    hints.ai_flags = AI_ADDRCONFIG | AI_NUMERICSERV;
+    hints.ai_flags = AI_ADDRCONFIG;
+#ifdef  AI_NUMERICSERV
+    hints.ai_flags |= AI_NUMERICSERV;
+#endif
     hints.ai_family = AF_UNSPEC;
     hints.ai_socktype = SOCK_STREAM;
     hints.ai_protocol = IPPROTO_TCP;
@@ -406,6 +409,7 @@ static int usbredirtestclient_cmdline_ctrl(void)
             }
             if (!arg || *endptr != '\0') {
                 printf("Missing or invalid data byte(s)\n");
+                free(data);
                 return 0;
             }
         }
